# Copyright 2017 by Bearstech <py@bearstech.com>
#
# This file is part of nuka.
#
# nuka is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nuka is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nuka. If not, see <http://www.gnu.org/licenses/>.

from asyncio import subprocess
from asyncio import streams
import asyncio
import random
import socket
import time
import zlib
import os

import asyncssh
import asyncssh.misc

import nuka
from nuka import utils

DEFAULT_LIMIT = streams._DEFAULT_LIMIT

asyncssh_connections = {}
asyncssh_connections_tasks = {}
asyncssh_keypairs = []
asyncssh_known_hosts = []


class BaseProcess:

    async def send_message(self, message, drain=True):
        utils.proto_dumps_std(message, self.stdin)
        if drain:
            return self.stdin.drain()

    async def next_message(self):
        try:
            self.read_task = self._loop.create_task(self.stdout.readline())
            content_type = await self.read_task
            self.read_task = self._loop.create_task(self.stdout.readline())
            content_length = await self.read_task
        except asyncio.CancelledError:
            raise
        else:
            headers = (content_type or b'') + (content_length or b'')
            self.read_task = None
            try:
                if isinstance(content_type, bytes):
                    content_type = content_type.decode('utf8')
                content_type = content_type.split(':')[1].strip()
                if isinstance(content_length, bytes):
                    content_length = content_length.decode('utf8')
                content_length = int(content_length.split(':')[1].strip())
            except IndexError:
                stdout = await self.stdout.read()
                if stdout or headers:
                    print((headers, stdout))
                stderr = await self.stderr.read()
                if stderr:
                    stderr = stderr.decode('utf8')
                    err = stderr.lower()
                    exc = OSError
                    if 'could not resolve hostname' in err:
                        exc = LookupError
                    elif 'host key verification failed' in err:
                        exc = LookupError
                    elif 'permission denied' in err:
                        exc = LookupError
                    exc = exc(stderr, self.host)
                    if isinstance(exc, LookupError):
                        self.host.fail(exc)
                    raise exc

                else:
                    raise ValueError((content_length, stderr))
            data = b''
            while len(data) < content_length:
                try:
                    coro = self.stdout.read(content_length - len(data))
                    self.read_task = self._loop.create_task(coro)
                    data += await self.read_task
                except asyncio.CancelledError:
                    raise
            if content_type == 'zlib':
                data = zlib.decompress(data)
            data = data.decode('utf8')
            try:
                data = utils.json.loads(data)
            except ValueError:
                raise ValueError(data)
            self.host.log.debug5(data)
            if data.get('message_type') == 'exit':
                await self.wait()
                duration = time.time() - self.start
                latency = duration - data['meta']['remote_time']
                self.host.add_time(
                    type='process', cmd=self.cmd,
                    start=self.start, time=duration, latency=latency,
                    task=self.task, meta=data['meta'])
            return data

    async def exit(self):
        if self.returncode is None:
            try:
                await self.wait()
            except asyncio.CancelledError:
                pass
        if getattr(self, '_transport', None):
            for i in (0, 1, 2):
                self._transport.get_pipe_transport(i).close()
        else:
            self.stdin.close()
        self.host.free_session_slot()
        self.host._processes.pop(id(self), None)


class Process(subprocess.Process, BaseProcess):

    def __init__(self, transport, protocol, host, task, cmd, start):
        super().__init__(transport, protocol, host.loop)
        self.host = host
        self.task = task
        self.cmd = cmd
        self.start = start
        self.read_task = None


class SSHClient(asyncssh.SSHClient):

    def __init__(self, uid):
        self.uid = uid
        self.start = time.time()

    def connection_made(self, *args, **kwargs):
        now = time.time()
        asyncssh_connections[self.uid]['connect'] = now - self.start
        self.start = now

    def auth_completed(self, *args, **kwargs):
        asyncssh_connections[self.uid]['auth_time'] = time.time() - self.start


class SSHClientProcess(asyncssh.SSHClientProcess, BaseProcess):

    def __init__(self, host, task, cmd, start):
        super().__init__()
        self._encoding = None
        self.host = host
        self.task = task
        self.cmd = cmd
        self.start = start
        self.read_task = None

    @property
    def returncode(self):
        return self.exit_status


async def get_keys(loop):
    if not asyncssh_keypairs:
        asyncssh_keypairs[:] = [None]
        agent_path = os.environ.get('SSH_AUTH_SOCK', None)
        agent = await asyncssh.connect_agent(
            agent_path, loop=loop)
        asyncssh_keypairs[:] = await agent.get_keys()
        agent.close()
    elif asyncssh_keypairs == [None]:
        while asyncssh_keypairs == [None]:
            await asyncio.sleep(.1)
    return asyncssh_keypairs[:]


def get_known_hosts(filename):
    if not asyncssh_known_hosts:
        known_hosts = asyncssh.known_hosts.read_known_hosts(filename)
        asyncssh_known_hosts[:] = [known_hosts]
    return asyncssh_known_hosts[0]


async def delay_connection(uid, loop):
    if uid not in asyncssh_connections:
        delay = nuka.config['connections']['delay']
        now = time.time()
        asyncssh_connections[uid] = {'now': now, 'timeouts': 0}
        if delay:
            delay = (delay * len(asyncssh_connections))
            min_time = asyncssh_connections.setdefault('time', now) + delay
            delay = min_time - now
            if delay:
                asyncssh_connections[uid]['delay'] = delay
                await asyncio.sleep(delay, loop=loop)


async def create(cmd, host, task=None):
    host.log.debug5(cmd)
    loop = host.loop

    start = time.time()

    if not cmd[0].startswith('ssh') or nuka.cli.args.ssh:
        def protocol_factory():
            return subprocess.SubprocessStreamProtocol(
                loop=loop, limit=DEFAULT_LIMIT)

        await host.acquire_session_slot()
        transport, protocol = await loop.subprocess_exec(
            protocol_factory,
            *cmd,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            preexec_fn=os.setpgrp,
            close_fds=True)

        proc = Process(transport, protocol, host, task, cmd, start)
    else:
        def protocol_factory():
            return SSHClientProcess(host, task, cmd, start)

        # retrieve params from ssh command
        tmp_cmd = cmd[:]
        ssh_cmd = tmp_cmd.pop()
        hostname = tmp_cmd.pop()
        agent_forwarding = False
        known_hosts = ()
        attempts = 1
        timeout = 924  # Default TCP Timeout on debian
        while tmp_cmd:
            v = tmp_cmd.pop(0)
            if v == '-l':
                username = tmp_cmd.pop(0)
            elif v == '-p':
                port = tmp_cmd.pop(0)
            elif v in ('-A', '-oForwardAgent=yes'):
                agent_forwarding = True
            elif v == '-oStrictHostKeyChecking=no':
                known_hosts = None
            elif v.startswith('-oConnectionAttempts'):
                attempts = int(v.split('=', 1)[1].strip())
            elif v.startswith('-oConnectTimeout'):
                timeout = int(v.split('=', 1)[1].strip())

        if known_hosts is not None:
            filename = os.path.expanduser('~/.ssh/known_hosts')
            if os.path.isfile(filename):
                try:
                    known_hosts = get_known_hosts(filename)
                except ValueError as e:
                    host.fail(e)
                    msg = ' '.join(e.args)
                    nuka.run_vars['exit_message'] = 'FATAL: ' + msg
                    return

        uid = (username, host)
        conn = asyncssh_connections.get(uid, {}).get('conn')
        if conn is None:
            exc = None
            client_keys = await get_keys(loop)
            asyncssh_connections_tasks[uid] = loop.create_task(
                delay_connection(uid, loop)
            )
            try:
                await asyncssh_connections_tasks[uid]
            except asyncio.CancelledError as e:
                exc = LookupError(OSError('sigint'), host)
                host.fail(exc)
                return

            for i in range(1, attempts + 1):
                host.log.debug5('open connection {0}/{1} at {2}'.format(
                                i, attempts, time.time()))
                try:
                    if nuka.run_vars['sigint']:
                        exc = LookupError(OSError('sigint'), host)
                        break
                    asyncssh_connections_tasks[uid] = loop.create_task(
                        asyncssh.create_connection(
                            lambda: SSHClient(uid),
                            hostname, port,
                            username=username,
                            known_hosts=known_hosts,
                            agent_forwarding=agent_forwarding,
                            client_keys=client_keys,
                            loop=loop,
                            )
                        )
                    conn, client = await asyncio.wait_for(
                            asyncssh_connections_tasks[uid],
                            timeout=timeout, loop=loop)
                    break
                except asyncio.CancelledError as e:
                    exc = LookupError(OSError('sigint'), host)
                    break
                except asyncio.TimeoutError as e:
                    timeouts = asyncssh_connections[uid]['timeouts']
                    asyncssh_connections[uid]['timeouts'] = timeouts + 1
                    if i == attempts:
                        host.log.error('TimeoutError({0}) exceeded '.format(
                            timeout, i, attempts))
                        exc = LookupError(e, host)
                    else:
                        asyncssh_connections_tasks[uid] = loop.create_task(
                            asyncio.sleep(1 + random.random(), loop=loop)
                        )
                        try:
                            await asyncssh_connections_tasks[uid]
                        except asyncio.CancelledError as e:
                            exc = LookupError(e, host)
                            break
                except (OSError, socket.error) as e:
                    exc = LookupError(e, host)
                    break

            if exc is not None:
                host.fail(exc)
                raise exc

            asyncssh_connections[uid]['conn'] = conn
        await host.acquire_session_slot()
        chan, proc = await conn.create_session(
                protocol_factory, ssh_cmd, encoding=None)
        await proc.redirect(asyncssh.PIPE, asyncssh.PIPE, asyncssh.PIPE,
                            DEFAULT_LIMIT)
    host._processes[id(proc)] = proc
    loop.create_task(proc.exit())

    return proc


def close_connections():
    for d in asyncssh_connections.values():
        if isinstance(d, dict):
            if 'conn' in d:
                d['conn'].close()


def close_awaiting_tasks():
    for task in asyncssh_connections_tasks.values():
        if not task.done():
            task.cancel()



# -*- coding: utf-8 -*-
import os
import pytest
import nuka
from nuka.tasks import http


@pytest.mark.asyncio
async def test_fetch(host):
    res = await http.fetch('http://bearstech.com')
    assert res.dst == os.path.join(nuka.config['remote_tmp'], 'bearstech.com')
    res = await http.fetch('http://bearstech.com', dst='/tmp/bt.com')
    assert res.dst == '/tmp/bt.com'



# -*- coding: utf-8 -*-
import os
import codecs

DEBIAN_RELEASES = (
    (u'9.', u'stretch'),
    (u'8.', u'jessie'),
    (u'7.', u'squeeze')
)


def update_inventory(inventory):
    infos = {'name': None, 'version': None, 'release': None}
    if os.path.isfile('/etc/debian_version'):
        infos['name'] = 'debian'
        with codecs.open('/etc/debian_version', 'r', 'utf8') as fd:
            version = fd.read().strip()
            infos['version'] = version
        for ver, name in DEBIAN_RELEASES:
            if version.startswith(ver):
                infos['release'] = name
                break
    inventory['os'] = infos



# -*- coding: utf-8 -*-
from nuka.tasks import service
from nuka.tasks import apt
import pytest
import os


pytestmark = [
    pytest.mark.skipif(
        'centos' in os.environ['ENV_NAME'], reason='exclude centos'),
]


@pytest.mark.asyncio
async def test_01_install_services(host):
    await apt.install(['rsync'])


@pytest.mark.asyncio
async def test_start_doc(host):
    await service.start('rsync')


@pytest.mark.asyncio
async def test_restart_doc(host):
    await service.restart('rsync')


@pytest.mark.asyncio
async def test_stop_doc(host):
    await service.stop('rsync')


@pytest.mark.asyncio
async def test_service(host):
    assert await service.stop('rsync')
    assert await service.stop('rsync')
    assert await service.restart('rsync')
    assert await service.start('rsync')
    assert await service.stop('rsync')
    assert await service.start('rsync')
    assert await service.restart('rsync')


@pytest.mark.asyncio
async def test_service_diff(host, diff_mode):
    assert await service.stop('rsync')
    assert await service.start('rsync')



# -*- coding: utf-8 -*-
import subprocess
import sys

from nuka.hosts.base import Host
import nuka


class Vagrant(Host):
    """A host configured using ``vagrant ssh-config``"""

    def __init__(self, **kwargs):
        kwargs['hostname'] = 'default'
        super().__init__(**kwargs)
        self.port = None
        self.config_file = '/tmp/nuka_vagrant'
        try:
            subprocess.check_call(
                'vagrant ssh-config > ' + self.config_file,
                shell=True)
        except subprocess.CalledProcessError:
            print('VM is down. Please run vagrant up...')
            sys.exit(1)

    def wraps_command_line(self, cmd, **kwargs):
        ssh_user = kwargs.get('switch_ssh_user')
        if ssh_user is None:
            # we use the main user account
            switch_user = kwargs.get('switch_user') or 'root'
            if switch_user != 'root':
                if switch_user != self.vars['user']:
                    # we have to use sudo
                    args = (switch_user, cmd)
                    if self.use_sudo:
                        cmd = '{sudo} -u {0} {1}'.format(*args, **nuka.config)
                    else:
                        cmd = '{su} -c "{1}" {0}'.format(*args, **nuka.config)
            elif self.use_sudo:
                cmd = '{sudo} {0}'.format(cmd, **nuka.config)

        if ssh_user is None:
            ssh_user = self.vars['user']

        ssh_cmd = ['ssh', '-F', self.config_file]
        ssh_cmd.extend(nuka.config['ssh']['options'] + ['-l', ssh_user])
        if self.port:
            ssh_cmd.extend(['-p', self.port])
        ssh_cmd.extend([self.hostname, cmd])
        return ssh_cmd



# -*- coding: utf-8 -*-
import codecs
from nuka.task import Task


class timezone(Task):

    def __init__(self, tz=None, **kwargs):
        # ensure we have a name to get a better repr() in logs
        kwargs.setdefault('name', tz)
        super(timezone, self).__init__(tz=tz, **kwargs)

    def do(self):
        """do the job: change the timezone file if needed"""
        tz = self.args['tz']
        changed = False
        with codecs.open('/etc/timezone', 'r', 'utf8') as fd:
            current_tz = fd.read().strip()
        if current_tz != tz:
            changed = True
            with codecs.open('/etc/timezone', 'w', 'utf8') as fd:
                current_tz = fd.write(tz + '\n')
        # we must return a dictionary with at least a return code and
        # the change state
        return dict(rc=0, changed=changed)

    def diff(self):
        """generate diff between actual state and task value.
        Implementing this method is not required but recommended"""
        tz = self.args['tz']
        with codecs.open('/etc/timezone', 'r', 'utf8') as fd:
            current_tz = fd.read().strip()
        diff = ''
        if current_tz != tz:
            diff = self.text_diff(current_tz, tz)
        return dict(diff=diff)



# Copyright 2017 by Bearstech <py@bearstech.com>
#
# This file is part of nuka.
#
# nuka is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nuka is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nuka. If not, see <http://www.gnu.org/licenses/>.

import logging
import sys
import os

import nuka
from nuka.utils import LOG
from nuka.utils import CHANGED


class HostLogger(logging.Logger):
    """Logger for a host that can log to a file and stdout at different
    levels"""

    def __init__(self, host):
        self.host = host
        stream_level = nuka.config['log']['levels']['stream_level']
        file_level = nuka.config['log']['levels']['file_level']
        super().__init__(host, max(stream_level, file_level))
        if stream_level is not None:
            if nuka.config['log']['quiet']:
                filename = nuka.config['log']['stdout']
                self.addHandler(HostFileHandler(host, level=stream_level,
                                                filename=filename))
            else:
                self.addHandler(HostStreamHandler(host, level=stream_level))
        self.addHandler(HostFileHandler(host, level=file_level))

    def changed(self, *args, **kwargs):
        self.log(CHANGED, *args, **kwargs)

    def __call__(self, msg, *args, **kwargs):
        self.log(LOG, msg, *args, **kwargs)

    def _extra_logs(self, level, i, *args, **kwargs):
        if nuka.cli.args.verbose > i:
            self.log(level, *args, **kwargs)

    def debug2(self, *args, **kwargs):
        self._extra_logs(logging.DEBUG, 2, *args, **kwargs)

    def debug3(self, *args, **kwargs):
        self._extra_logs(logging.DEBUG, 3, *args, **kwargs)

    def debug4(self, *args, **kwargs):
        self._extra_logs(logging.DEBUG, 4, *args, **kwargs)

    def debug5(self, *args, **kwargs):
        self._extra_logs(logging.DEBUG, 5, *args, **kwargs)

    def exception5(self, *args, **kwargs):
        if nuka.cli.args.verbose > 5:
            self.exception(*args, **kwargs)


class HostFileHandler(logging.FileHandler):

    __rollover = False

    def __init__(self, host, level, filename=None):
        self.host = host
        logdir = nuka.config['log']['dirname']
        if filename:
            self.filename = filename
            if not self.__rollover:
                # if filename is specified, rollover only once during run time
                self.__class__.__rollover = True
                self.rollover()
        else:
            self.filename = os.path.join(logdir, '{0}.log'.format(host))
            self.rollover()
        if not os.path.isdir(os.path.dirname(self.filename)):
            os.makedirs(os.path.dirname(self.filename))
        super().__init__(self.filename, 'w')
        fmt = nuka.config['log']['formats']['default']
        self.setFormatter(logging.Formatter(fmt))
        self.setLevel(level)

    def rollover(self):
        for i in range(9 - 1, 0, -1):
            sfn = "%s.%d" % (self.filename, i)
            dfn = "%s.%d" % (self.filename, i + 1)
            if os.path.exists(sfn):
                if os.path.exists(dfn):
                    os.remove(dfn)
                os.rename(sfn, dfn)
        dfn = self.filename + ".1"
        if os.path.exists(dfn):
            os.remove(dfn)
        if os.path.exists(self.filename):
            os.rename(self.filename, dfn)


class HostStreamHandler(logging.StreamHandler):

    reset = "\033[0m"

    def __init__(self, host, level):
        self.host = host
        super().__init__(sys.stdout)
        fmt = nuka.config['log']['formats']['host'].format(host)
        self.setFormatter(logging.Formatter(fmt))
        self.setLevel(level)

    def format(self, record):
        res = super().format(record)
        col = nuka.config['log']['colors'].get(record.levelno, '')
        res = res.split('\n', 1)
        if len(res) > 1:
            res = ''.join([col, res[0] + self.reset, '\n'] + res[1:])
        else:
            res = col + res[0] + self.reset
        return res



# -*- coding: utf-8 -*-
"""
services related tasks
"""
from nuka.task import Task


class start(Task):
    """ensure a service is started"""

    def __init__(self, name=None, **kwargs):
        super(start, self).__init__(name=name, **kwargs)

    def do(self):
        res = self.sh(['service', self.args['name'], 'status'], check=False)
        if res['rc'] != 0:
            res = self.sh(['service', self.args['name'], 'start'])
        else:
            res['changed'] = False
        return res

    def diff(self):
        res = self.sh(['service', self.args['name'], 'status'], check=False)
        diff = ''
        if res['rc'] != 0:
            diff = self.texts_diff('', self.args['name'])
        return dict(diff=diff)


class restart(Task):
    """restart one or more service"""

    diff = False

    def __init__(self, services=None, **kwargs):
        if isinstance(services, str):
            services = [services]
        kwargs.update(
            name=', '.join(services),
            services=services)
        super(restart, self).__init__(**kwargs)

    def do(self):
        for service in self.args['services']:
            res = self.sh(['service', service, 'restart'])
        return res


class reload(Task):
    """reload one or more service"""

    diff = False

    def __init__(self, services=None, **kwargs):
        if isinstance(services, str):
            services = [services]
        kwargs.update(
            name=', '.join(services),
            services=services)
        super(reload, self).__init__(**kwargs)

    def do(self):
        for service in self.args['services']:
            res = self.sh(['service', service, 'reload'])
        return res


class stop(Task):
    """ensure a service is stoped"""

    def __init__(self, name=None, **kwargs):
        super(stop, self).__init__(name=name, **kwargs)

    def do(self):
        return self.sh(['service', self.args['name'], 'stop'])

    def diff(self):
        res = self.sh(['service', self.args['name'], 'status'], check=False)
        if res['rc'] == 0:
            diff = self.lists_diff(['{0}\n'.format(self.args['name'])], [])
        else:
            diff = ''
        return dict(diff=diff)



#!../bin/python
from os.path import (join, basename)
import tempfile

import nuka
from nuka.hosts import DockerContainer
from nuka.tasks import (apt, user, virtualenv, shell, http, file)

nuka.config['templates'].append('/tmp')

kinto = DockerContainer('kinto')

WSGI_FILE = 'https://raw.githubusercontent.com/Kinto/kinto/master/app.wsgi'


async def install_kinto(host):
    kinto_user = user.create_www_user('kinto')
    await nuka.wait(
        kinto_user,
        apt.install([
            'apache2', 'libapache2-mod-wsgi',
            'build-essential', 'libffi-dev', 'libssl-dev',
            'python3-dev', 'redis-server'
        ], update_cache=3600))

    with open('/tmp/requirements.txt', 'wb') as fd:
        fd.write(b'kinto\n')
        fd.flush()
        await virtualenv.virtualenv(
            dst=kinto_user.home,
            requirements=fd.name,
            switch_user='kinto')

    binary = join(kinto_user.home, 'bin', 'kinto')
    config = join(kinto_user.home, 'config', 'kinto.ini')

    if not await file.exists(config):
        await shell.command([binary, 'init', '--backend=redis'],
                            switch_user='kinto')
    await shell.command([binary, 'migrate'],
                        switch_user='kinto')

    if not await file.exists('app.wsgi', switch_user='kinto'):
        await http.fetch(
            src=WSGI_FILE,
            dst='app.wsgi', switch_user='kinto')

    with tempfile.NamedTemporaryFile(suffix='.j2') as fd:
        fd.write(b'''
ServerName kinto.example.com

WSGIScriptAlias /         {{kinto_user.home}}/app.wsgi
WSGIPythonPath            {{kinto_user.home}}
SetEnv          KINTO_INI {{kinto_config}}

<Directory {{kinto_user.home}}>
  <Files app.wsgi>
    Require all granted
  </Files>
</Directory>
''')
        fd.flush()
        conf = await file.put([dict(
            src=basename(fd.name),
            dst='/etc/apache2/sites-available/kinto.conf')],
            ctx=dict(kinto_user=kinto_user, kinto_config=config))

    if conf.changed:
        await shell.command(['a2ensite', 'kinto.conf'])


nuka.run(install_kinto(kinto))



# -*- coding: utf-8 -*-
import importlib

_modules = (
    'zlib', 'psutils',
    'virtualenv', 'pip', 'setuptools',
    'requests',
    'lxml',
)


def update_inventory(inventory, modules=_modules):
    libs = inventory.setdefault('python_libs', {})
    for name in modules:
        try:
            importlib.import_module(name)
        except ImportError:
            libs[name] = False
        else:
            libs[name] = True



# -*- coding: utf-8 -*-
import os
import inspect
import importlib
from operator import attrgetter

from nuka.task import Task
from nuka import config


def main():
    # update README from docs/index.rst
    with open('docs/index.rst') as fd:
        lines = fd.readlines()
    new_lines = [(
        '.. Do not edit this file. It is generated from docs/index.rst.'
        ' See docs/utils.py\n\n'
    )]
    for line in lines:
        if 'examples/quickstart.py' in line:
            new_lines.extend(['\n', '::\n', '\n'])
            with open('examples/quickstart.py') as fd:
                new_lines.extend(['    ' + l for l in fd.readlines()])
        elif line == 'Quickstart\n':
            new_lines.extend([
                ('Full documentation is available at '
                 'http://doc.bearstech.com/nuka\n\n'),
                line,
            ])
        elif line == 'Index\n':
            break
        else:
            new_lines.append(line)
    with open('README.rst', 'w') as fd:
        fd.writelines(new_lines)

    # auto generated tasks
    for root, dirnames, filenames in os.walk('nuka/tasks'):
        for filename in filenames:
            if not filename.endswith('.py'):
                continue
            elif filename.startswith('__'):
                continue

            name = filename[:-3]

            if name in ('setup',):
                continue

            module = 'nuka.tasks.{0}'.format(name)
            mod = importlib.import_module(module)

            # get all tasks from module
            tasks = []
            for k in mod.__dict__:
                if not k.startswith('_'):
                    v = getattr(mod, k, None)
                    try:
                        if issubclass(v, Task) and v.__module__ == module:
                            tasks.append(v)
                    except TypeError:
                        pass

            # extract test code if any
            test_module = 'tests.tasks.test_{0}'.format(name)
            try:
                test = importlib.import_module(test_module)
            except ImportError:
                test = None
                test_module = None
            else:
                test_module = 'test_{0}'.format(name)

            for task in tasks:
                task.test_module = test_module
                task.test = getattr(
                    test, 'test_' + task.__name__ + '_doc', None)
                if task.test:
                    task.test_source = ''.join(
                        inspect.getsourcelines(task.test)[0][2:])

            # compile template and save it to docs/tasks
            ctx = dict(
                tasks=sorted(tasks, key=attrgetter('__name__')),
                module=module,
            )
            engine = config.get_template_engine()
            template = engine.get_template('sphinx/task_module.j2')
            with open('docs/tasks/{0}.rst'.format(name), 'w') as fd:
                fd.write(template.render(ctx))


if __name__ == '__main__':
    main()



#!../../bin/python
# -*- coding: utf-8 -*-
import nuka
from nuka.hosts import (DockerCompose, Cloud, Provider)
from nuka.tasks import (apt, shell, file, user, archive, service)

import mysql


nuka.cli.add_argument('--gce', action='store_true', default=False)
nuka.cli.add_argument('--destroy', action='store_true', default=False)
nuka.cli.parse_args()


if nuka.cli.args.gce:
    all_hosts = Cloud(Provider.GCE).from_compose()
else:
    all_hosts = DockerCompose()
    nuka.run(all_hosts.boot())


nuka.config['all_hosts'] = all_hosts
master = all_hosts['wordpress_master_1']
slave = all_hosts['wordpress_slave_1']
web = all_hosts['wordpress_web_1']


mysql_password = nuka.utils.secret('dsmfk;m#lkf!(').next()
db_password = nuka.utils.secret('q,sddsmfk;m#lkf!(').next()


async def install_wordpress(host):
    waits = [
        apt.install(['apache2', 'libapache2-mod-php5', 'ca-certificates',
                     'php5-mysql', 'mysql-client'], update_cache=3600)]

    u = await user.create_www_user('wp')
    wp_dir = u.home + '/wordpress'
    if not await file.exists(wp_dir):
        fetch = archive.untar(src='https://wordpress.org/latest.tar.gz',
                              dst=u.home,
                              switch_user=u.username)
        waits.append(fetch)
    await nuka.wait(waits)

    confs = await nuka.wait(
        file.put(
            [dict(src='wordpress/wp-config.php.j2',
                  dst=wp_dir + '/wp-config.php')],
            ctx=dict(database='wp', user='wp',
                     password=db_password,
                     hostname=master.private_ip,
                     secret=nuka.utils.secret('dsoi,mfjqm'),
                     debug=False,
                     ),
            switch_user=u.username),
        file.put(
            [dict(src='wordpress/apache.conf.j2',
                  dst='/etc/apache2/sites-available/wp.conf')],
            ctx=dict(document_root=wp_dir))
    )

    await nuka.wait(mysql.mysql_ready)
    if any([c.changed for c in confs]):
        await shell.command(['a2dissite', '000-default.conf'])
        await shell.command(['a2ensite', 'wp'])
        await service.restart('apache2')
    host.log('URL: http://' + host.hostname)


if nuka.cli.args.destroy:
    nuka.run(all_hosts.destroy())
else:
    nuka.run(
        mysql.install_master(
            master, mysql_password,
            db_name='wp', db_user='wp', db_password=db_password,
            slave=slave),
        mysql.install_slave(slave, mysql_password),
        install_wordpress(web),
    )



# -*- coding: utf-8 -*-
"""
command line related tasks
"""
from nuka import utils
from nuka.task import Task


class commands(Task):
    """run multiple command"""

    def __init__(self, cmds=None, **kwargs):
        kwargs.setdefault('name', cmds)
        super(commands, self).__init__(cmds=cmds, **kwargs)

    def do(self):
        for cmd in self.args['cmds']:
            kwargs = {}
            watch = self.args.get('watch')
            if watch:  # pragma: no cover
                kwargs['watcher'] = utils.default_watcher(delay=watch)
            res = self.sh(cmd, **kwargs)
        res.pop('stderr')
        return res


class command(Task):
    """run a task"""

    def __init__(self, cmd=None, **kwargs):
        kwargs.setdefault('name', cmd)
        super(command, self).__init__(cmd=cmd, **kwargs)

    def do(self):
        kwargs = {}
        watch = self.args.get('watch')
        if watch:  # pragma: no cover
            kwargs['watcher'] = utils.default_watcher(delay=watch)
        res = self.sh(self.args['cmd'], **kwargs)
        res.pop('stderr')
        return res


class shell(Task):
    """run a shell line (allow pipes)"""

    def __init__(self, cmd=None, **kwargs):
        kwargs.setdefault('name', cmd)
        super(shell, self).__init__(cmd=cmd, **kwargs)

    def do(self):
        kwargs = {'shell': True}
        watch = self.args.get('watch')
        if watch:  # pragma: no cover
            kwargs['watcher'] = utils.default_watcher(delay=watch)
        cmd = self.args['cmd']
        if isinstance(cmd, list):
            cmd = ' '.join(cmd)
        res = self.sh(cmd, **kwargs)
        res.pop('stderr')
        return res



# -*- coding: utf-8 -*-
import nuka
from nuka.hosts import DockerContainer

master = DockerContainer('master')
slave = DockerContainer('slave')

e = nuka.Event('master_ready')


async def my_tasks(host):
    if host is slave:
        await nuka.wait(e)
        host.log.warn('starting slave')
    else:
        host.log.warn('master ready')
        e.release()

nuka.run(
    my_tasks(slave),
    my_tasks(master),
)



# -*- coding: utf-8 -*-
"""
"""
import os
import tempfile
import posixpath
from nuka.task import Task


class fetch(Task):
    """fetch content from internet"""

    def __init__(self, src=None, dst=None, **kwargs):
        kwargs.setdefault('name', src)
        super(fetch, self).__init__(src=src, dst=dst, **kwargs)

    def fetch(self, dst, src):
        if os.path.isfile('/usr/bin/curl'):
            res = self.sh(['curl', '-so', dst, src])
        elif os.path.isfile('/usr/bin/wget'):
            res = self.sh(['wget', '-qO', dst, src])
        else:
            from nuka.utils import urlretrieve
            urlretrieve(src, dst)
            res = {'rc': 0}
        return res

    def do(self):
        src = self.args['src']
        dst = self.args['dst']
        if dst is None:
            filename = posixpath.split(src.strip('/'))[-1]
            if not filename:
                with tempfile.NamedTemporaryFile(prefix='http_fetch_') as fd:
                    dst = fd.name
            else:
                dst = os.path.join(tempfile.gettempdir(), filename)
        res = self.fetch(dst, src)
        res.update(dst=dst)
        return res



# -*- coding: utf-8 -*-
from nuka.tasks import apt
import pytest
import os

pytestmark = [
    pytest.mark.skipif(
        'centos' in os.environ['ENV_NAME'], reason='exclude centos'),
]


@pytest.mark.asyncio
async def test_source_doc(host):
    if 'wheezie' in host.hostname:
        n = 'wheezie'
    elif 'jessie' in host.hostname:
        n = 'jessie'
    else:
        n = 'stretch'
    src = 'deb http://apt.dockerproject.org/repo/ debian-{0} main'.format(n)
    res = await apt.source(
        name='docker',
        key='https://yum.dockerproject.org/gpg',
        src=src,
    )
    assert bool(res)
    src = 'deb http://deb.bearstech.com/debian {0}-bearstech main'.format(n)
    res = await apt.source(
        name='bearstech',
        key='https://deb.bearstech.com/bearstech-archive.gpg',
        src=src,
    )
    assert bool(res)


@pytest.mark.asyncio
async def test_debconf_set_selections_doc(host):
    res = await apt.debconf_set_selections(
        [('adduser', 'adduser/homedir-permission', 'true')]
    )
    assert bool(res)


@pytest.mark.asyncio
async def test_update_doc(host):
    res = await apt.update(cache=3600)
    assert bool(res)


@pytest.mark.asyncio
async def test_install_doc(host):
    res = await apt.install(['python'])
    assert bool(res)


@pytest.mark.asyncio
async def test_install_diff(host, diff_mode):
    with diff_mode:
        res = await apt.install(packages=['moreutils'])
        assert res.rc == 0
        assert '+moreutils\n' in res.res['diff'], res.res['diff']



# -*- coding: utf-8 -*-
import os
from setuptools import setup
from setuptools import find_packages

version = '0.3'


def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()


docker = ['docker-compose']
cloud = ['python-novaclient', 'apache-libcloud', 'PyCrypto']
test = ['pytest', 'pytest-asyncio', 'coverage']
full = ['ujson'] + docker + cloud + test


setup(
    name='nuka',
    version=version,
    description="provisioning tool focused on performance.",
    long_description=read('README.rst'),
    classifiers=[
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3.5',
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Console',
        ('License :: OSI Approved :: '
         'GNU General Public License v3 or later (GPLv3+)'),
        'Operating System :: POSIX',
        'Topic :: System :: Systems Administration',
    ],
    keywords='devops docker vagrant gce',
    license='GPLv3',
    author='Bearstech',
    author_email='py@bearstech.com',
    packages=find_packages(exclude=['docs', 'tests']),
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'pyaml',
        'jinja2',
        'uvloop',
        'asyncssh',
    ],
    extras_require={
        'full': ['tox'] + full,
        'speedup': ['ujson'],
        'docker': docker,
        'cloud': cloud,
        'test': full,
    },
    entry_points="""
    [pytest11]
    nuka = pytest_nuka
    """,
)



import os
import sys
import codecs
from nuka.task import Task
from nuka.utils import json
from nuka.utils import import_module


class setup(Task):

    @classmethod
    def get_inventory(self):
        modules = [m.split('=')[1] for m in sys.argv
                   if m.startswith('--inventory=')]
        modules.insert(0, 'nuka.inventory.python')

        inventory = {}
        done = set()
        for name in modules:
            if name not in done:
                done.add(name)
                mod = import_module(name)
                meth = getattr(mod, 'update_inventory', None)
                if meth is not None:
                    meth(inventory)
        return {'inventory': inventory}

    def do(self):
        dirname = os.path.dirname(sys.argv[0])
        cache = os.path.join(dirname, 'inventory.json')
        data = self.get_inventory()
        with codecs.open(cache, 'w') as fd:
            json.dump(data, fd)
        return data



# -*- coding: utf-8 -*-
import nuka
from nuka.hosts import DockerContainer
from nuka.hosts import Host
from nuka.tasks import shell

nuka.config['ssh']['extra_options'] = ['-C', '-oStrictHostKeyChecking=yes']

host1 = DockerContainer(hostname='debian_jessie')
host2 = Host(hostname='amandine.bearstech.com')


async def echo():
    print('echo')


async def ls(host):
    res = await shell.command(['ls'])
    return res


print(nuka.run(ls(host2)))
print(nuka.run(echo(), ls(host1), ls(host2)))



# Copyright 2017 by Bearstech <py@bearstech.com>
#
# This file is part of nuka.
#
# nuka is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nuka is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nuka. If not, see <http://www.gnu.org/licenses/>.
import tempfile
import logging
import os

import jinja2
import yaml

from nuka.utils import CHANGED
from nuka.gpg import FileSystemLoader


asyncio_logger = logging.getLogger('asyncio')


class Config(dict):

    def update_from_file(self, yaml_config):  # pragma: no cover
        for doc in yaml.load_all(yaml_config):
            for k, v in doc.items():
                if isinstance(v, dict):
                    self.setdefault(k, v)
                    self[k].update(v)
                else:
                    self[k] = v

    def finalize(self, args):
        if args.nuka_dir:
            self['nuka_dir'] = args.nuka_dir
        nuka_dir = self['nuka_dir']
        tempdir = args.tempdir or tempfile.mkdtemp(prefix='nuka')
        self['tmp'] = tempdir
        for dirname in (nuka_dir, tempdir):
            if not os.path.isdir(dirname):  # pragma: no cover
                os.makedirs(dirname)

        for name in ('log', 'reports', 'ssh'):
            for key, value in self[name].items():
                if isinstance(value, str):
                    value = value.format(nuka_dir=nuka_dir)
                    self[name][key] = value
                    if key == 'dirname':
                        if not os.path.isdir(value):  # pragma: no cover
                            os.makedirs(value)

        # set correct path in self
        opts = self['ssh']['options']
        opts.extend(self['ssh'].get('extra_options', []))
        self['ssh']['options'] = []
        for o in opts:
            self['ssh']['options'].append(o.format(**config['ssh']))

        self['remote_dir'] = os.path.join(tempdir, 'nuka')
        self['remote_tmp'] = os.path.join(tempdir, 'tmp')
        self['script'] = os.path.join(config['remote_dir'], 'script.py')

        if args.verbose == 1:  # pragma: no cover
            self['log']['levels']['stream_level'] = logging.INFO
            self['log']['levels']['file_level'] = logging.INFO
            self['log']['levels']['remote_level'] = logging.WARN
        elif args.verbose == 2:  # pragma: no cover
            self['log']['levels']['stream_level'] = logging.INFO
            self['log']['levels']['file_level'] = logging.DEBUG
            self['log']['levels']['remote_level'] = logging.INFO
        elif args.verbose > 2:
            self['log']['levels']['stream_level'] = logging.DEBUG
            self['log']['levels']['file_level'] = logging.DEBUG
            self['log']['levels']['remote_level'] = logging.DEBUG

        if args.verbose < 3:
            asyncio_logger.setLevel(logging.CRITICAL)

        if args.quiet or 'quiet' not in self['log']:
            self['log']['quiet'] = args.quiet

        if args.connections_delay or 'delay' not in self['connections']:
            self['connections']['delay'] = args.connections_delay

    def get_template_engine(self):
        engine = self.get('template_engine')
        if engine is None:
            templates = self['templates']
            dirname = os.path.join(os.getcwd(), 'templates')
            if os.path.isdir(dirname):  # pragma: no cover
                if dirname not in templates:
                    templates.insert(0, dirname)
            elif os.getcwd() not in templates:
                templates.insert(0, os.getcwd())
            loader = jinja2.ChoiceLoader([
                FileSystemLoader(p) for p in templates
            ] + [jinja2.PackageLoader('nuka')])
            self['template_engine'] = jinja2.Environment(
                loader=loader,
                undefined=jinja2.StrictUndefined,
                keep_trailing_newline=True,
                autoescape=False,
            )
        return self['template_engine']


config = Config()
config['id'] = id(config)
config['testing'] = 'TESTING' in os.environ
config['templates'] = []

config['nuka_dir'] = '.nuka'

config['inventory_modules'] = []

config['sudo'] = 'sudo'
config['su'] = 'su -l'

config['docker'] = {
    'use_api': False,
}

config['ssh'] = {
    'dirname': '{nuka_dir}/ssh',
    'options': [
       '-oControlMaster=auto',
       '-oControlPersist=300s',
       '-oControlPath={dirname}/%r@%h:%p',
    ],
}
config['connections'] = {'delay': .2}
config['log'] = {
    'dirname': '{nuka_dir}/logs',
    'stdout': '{nuka_dir}/logs/stdout.log',
    'formats': {
        'default': '%(levelname)-5.5s: %(message)s',
        'host': '%(levelname)-5.5s:{0.name:15.15}: %(message)s',
    },
    'levels': {
        'stream_level': CHANGED,
        'file_level': logging.INFO,
        'remote_level': logging.INFO,
    },
    'colors': {
        logging.ERROR: "\033[01;31m",
        logging.CRITICAL: "\033[01;31m",
        logging.WARN: "\033[01;35m",
        CHANGED: "\033[01;35m",
    }
}
config['reports'] = {
    'dirname': '{nuka_dir}/reports',
}



#!../bin/python
# -*- coding: utf-8 -*-
import nuka
from nuka.hosts import Cloud
from nuka.tasks import shell

nuka.config['gce'] = {
    'driver': '~/.gce/nuka.json',
    'user': 'gawel',
    'create_node_args': {
        'size': 'f1-micro',
        'image': 'debian-8-jessie-v20161215',
        'location': 'europe-west1-d',
        'ex_tags': ['nuka'],
        'ex_disk_auto_delete': True,
        'ex_service_accounts': [{
            'scopes': [
                'https://www.googleapis.com/auth/cloud.useraccounts.readonly',
                'https://www.googleapis.com/auth/devstorage.read_only',
                'https://www.googleapis.com/auth/logging.write',
                'https://www.googleapis.com/auth/monitoring.write',
                'https://www.googleapis.com/auth/service.management.readonly',
                'https://www.googleapis.com/auth/servicecontrol'
             ],
        }]
    },
}

nuka.config['ssh'].update({
    'extra_options': ['-C', '-oStrictHostKeyChecking=no'],
    'keysfile': '~/.ssh/authorized_keys',
})


cloud = Cloud('gce')
host = cloud.get_or_create_node('myhost')


async def my_tasks(host):
    await shell.command(['ls'])

nuka.cli.add_argument('--destroy', action='store_true', default=False)
nuka.cli.parse_args()

if nuka.cli.args.destroy:
    nuka.run(cloud.destroy())
else:
    nuka.run(my_tasks(host))



# -*- coding: utf-8 -*-
"""
apache related tasks
"""
import os
from nuka.task import Task


class a2ensite(Task):
    """a2ensite"""

    def __init__(self, names=None, **kwargs):
        super(a2ensite, self).__init__(names=names, **kwargs)

    def pre_process(self):
        args = self.args
        class_name = self.__class__.__name__
        if class_name.startswith('a2en'):
            enable = True
            type = class_name[4:]
        else:
            enable = False
            type = class_name[5:]
        type = type if type == 'conf' else type + 's'
        args.update(
            binary=class_name,
            enable=enable,
            type=type,
        )
        elements = []
        names = args['names']
        if isinstance(names, str):
            names = [names]
        for name in names:
            args['name'] = name
            filename = (
                '/etc/apache2/{type}-enabled/{name}'.format(**args))
            if type == 'mods':
                filename += '.load'
            elif type == 'conf':
                filename += '.conf'
            elements.append((name, filename))
        args.update(name=', '.join(names), elements=elements)

    def do(self):
        diffs = self.get_diffs()
        if diffs:
            return self.sh([self.args['binary']] + diffs)
        else:
            return dict(rc=0, changed=False)

    def get_diffs(self):
        diffs = []
        for name, filename in self.args['elements']:
            if self.args['enable']:
                if not os.path.isfile(filename):
                    diffs.append(name)
            elif os.path.isfile(filename):
                    diffs.append(name)
        return diffs

    def diff(self):
        diffs = self.get_diffs()
        diffs = [n + '\n' for n in diffs]
        if self.args['enable']:
            diff = self.lists_diff([], diffs)
        else:
            diff = self.lists_diff(diffs, [])
        return dict(rc=0, diff=diff)


class a2dissite(a2ensite):
    """a2dissite"""


class a2enmod(a2ensite):
    """a2enmod"""


class a2dismod(a2ensite):
    """a2dismod"""


class a2enconf(a2ensite):
    """a2enconf"""


class a2disconf(a2ensite):
    """a2disconf"""



# -*- coding: utf-8 -*-
"""
git related tasks
"""
import os
import logging as log
from nuka.task import Task


class git(Task):
    """git clone/fetch"""

    def __init__(self, src=None, dst=None,
                 branch=None, tag=None, **kwargs):
        kwargs.setdefault('name', src)
        kwargs.update(src=src, dst=dst, branch=branch, tag=tag)
        super(git, self).__init__(**kwargs)

    def pre_process(self):
        if self.args['branch'] and self.args['tag']:
            raise RuntimeError('Do not use both tag and branch')
        if self.args['branch'] is None and self.args['tag'] is None:
            self.args['branch'] = 'master'
        if self.args['tag']:
            self.args['depth'] = 1

    def current_ref(self):
        ref = self.sh(['git', 'rev-parse', 'HEAD'], check=False)
        if ref['rc'] == 0:
            ref = ref['stdout'].strip()
            return ref

    def do(self):
        src = self.args['src']
        dst = self.args['dst']
        branch = self.args['branch']
        tag = self.args['tag']
        ref = branch or tag
        cmd = ['git']
        if os.path.isdir(dst):
            old_ref = self.current_ref()
            os.chdir(dst)
            cmd.append('fetch')
            if tag:
                cmd.append('--depth=1')
            cmd.extend([src, ref])
        else:
            old_ref = None
            cmd += ['clone']
            if tag:
                cmd.extend([
                    '--depth=1', '--single-branch',
                    '-b', tag, '--depth=1'])
            cmd.extend([src, dst])
        log.info(cmd)
        self.sh(cmd)
        os.chdir(dst)
        if tag:
            self.sh(['git',  'fetch', '--tags'])
        res = self.sh(['git', 'checkout',  ref])
        res['changed'] = old_ref != self.current_ref()
        return res



# -*- coding: utf-8 -*-
import os
import sys
import pytest
import subprocess

import nuka
from nuka import utils
from nuka import config
from nuka.remote.task import Task


class command(Task):

    def do(self):
        return self.sh(self.args['cmd'])


script = os.path.join(
    os.path.dirname(nuka.__file__),
    'remote', 'script.py')


@pytest.fixture
def script_process(request):
    def _script_process(*args):
        p = subprocess.Popen(
            [sys.executable, script] + list(args),
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        return p
    return _script_process


@pytest.fixture
def remote_stdin(request):
    def _remote_stdin(**kwargs):
        cmd = kwargs.pop('cmd', ['echo', 'bear'])
        stdin = dict(
            task=('tests.test_remote_script', 'command'),
            remote_tmp=config['remote_tmp'],
            switch_user=None,
            args={'cmd': cmd, 'name': ''},
            check_mode=False,
            diff_mode=False,
            log_level=config['log']['levels']['remote_level'])
        stdin.update(kwargs)
        std = utils.io.BytesIO()
        utils.proto_dumps_std(stdin, std)
        std.seek(0)
        return std.read()
    yield _remote_stdin


def test_simple_script(script_process, remote_stdin):
    p = script_process()
    data = remote_stdin()
    stdout, stderr = p.communicate(data)
    assert not stderr
    data = utils.proto_loads_std(stdout)
    assert data['rc'] == 0


def test_kill_script(script_process, remote_stdin):
    p = script_process()
    data = remote_stdin(cmd=['sleep', '5'])
    data += utils.proto_dumps(dict(signal='SIGINT'))
    stdout, stderr = p.communicate(data)
    assert p.returncode == 0
    assert not stderr
    assert stdout
    data = utils.proto_loads_std(stdout)
    assert data['rc'] == 1, data
    assert data['signal'] == 'SIGINT'



# -*- coding: utf-8 -*-
import sys


def update_inventory(inventory):
    try:
        import zlib  # NOQA
        zlib_avalaible = True
    except ImportError:  # pragma: no cover
        zlib_avalaible = False
    data = {
        'executable': sys.executable,
        'python_version': list(sys.version_info),
        'zlib_available': zlib_avalaible,
    }
    inventory['python'] = data



# -*- coding: utf-8 -*-
import tarfile
import time
import sys
import os
import io

import nuka


def exclude_pyc(info):
    if info.name.endswith(('.py', 'tasks', 'inventory')):
        return info


def build_archive(extra_classes=[], mode='x:gz'):
    """build a tarball with required scripts and python modules"""
    try:
        return build_archive.archives[mode]
    except KeyError:

        modules = nuka.config['inventory_modules'][:]
        modules.extend([klass.__module__ for klass in extra_classes])

        dirnames = set()
        filenames = set()
        for module in modules:
            if module.startswith('nuka.'):
                continue
            mod = sys.modules[module]
            mod_path = module.replace('.', '/') + '.py'
            filenames.add((mod.__file__, mod_path))
            dirname = mod_path
            while os.sep in dirname:
                dirname = os.path.dirname(mod_path)
                dirnames.add(dirname)

        nuka_dir = os.path.dirname(nuka.__file__)
        fd = io.BytesIO()
        with tarfile.open(fileobj=fd, mode=mode) as tfd:
            for dirname in dirnames:
                filename = dirname + '/__init__.py'
                if filename not in filenames:
                    tarinfo = tarfile.TarInfo(name=filename)
                    tarinfo.size = len(b'')
                    tarinfo.mtime = time.time()
                    tfd.addfile(tarinfo, io.BytesIO(b''))
            for src, dst in filenames:
                tfd.add(src, dst)
            tfd.add(os.path.join(nuka_dir, 'inventory'),
                    'nuka/inventory', filter=exclude_pyc)
            tfd.add(os.path.join(nuka_dir, 'tasks'),
                    'nuka/tasks', filter=exclude_pyc)
            tfd.add(os.path.join(nuka_dir, 'remote/task.py'),
                    'nuka/task.py')
            tarinfo = tarfile.TarInfo(name='nuka/__init__.py')
            tarinfo.size = len(b'')
            tarinfo.mtime = time.time()
            tfd.addfile(tarinfo, io.BytesIO(b''))
            tfd.add(os.path.join(nuka_dir, 'utils.py'),
                    'nuka/utils.py')
            tfd.add(os.path.join(nuka_dir, 'remote/script.py'),
                    'script.py')

            filenames = tfd.getnames()

        fd.seek(0)
        build_archive.archives[mode] = fd.read()

        if nuka.cli.args.verbose > 6:
            print('tarfile({0}ko): \n - {1}'.format(
                int(len(build_archive.archives[mode]) / 1024),
                '\n - '.join(filenames)))

        return build_archive.archives[mode]


build_archive.archives = {}



# -*- coding: utf-8 -*-
import pytest
from nuka.tasks import file
from nuka.tasks import user
from nuka.tasks import shell


@pytest.mark.asyncio
async def test_create_user(host, diff_mode):
    diff_mode(False)
    assert (await user.delete_user('test_user1'))

    with diff_mode:
        res = await user.delete_user(username='test_user1')
        assert '-test_user1' not in res.res['diff']

        res = await user.create_user(username='test_user1')
        assert '+test_user1' in res.res['diff']

    assert (await user.create_user(username='test_user1'))

    with diff_mode:
        res = await user.delete_user(username='test_user1')
        assert '-test_user1' in res.res['diff']

        res = await user.create_user(username='test_user1')
        assert '+test_user1' not in res.res['diff']

    assert (await user.delete_user('test_user1'))
    assert not (await file.exists('/home/test_user1'))
    assert (await user.create_user(username='test_user1'))
    assert (await file.exists('/home/test_user1'))
    res = await shell.shell(['whoami'], switch_user='test_user1')
    assert res.stdout.strip() == 'test_user1'

    assert (await user.delete_user('test_user1'))
    assert not (await file.exists('/home/test_user1'))


@pytest.mark.asyncio
async def test_create_user_doc(host):
    await user.create_user('myuser')


@pytest.mark.asyncio
async def test_authorized_keys_doc(host):
    await user.create_user('myuser')
    await user.authorized_keys(
        username='myuser', keysfile='~/.ssh/authorized_keys')


@pytest.mark.asyncio
async def test_delete_user_doc(host):
    await user.delete_user('myuser')
